-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;
use work.keccak_pkg.all;
	
-- implementation of Keccak round: there are two basic architectures of Keccak function
-- Guido Bertoni implementation of Keccak round - streightforward
-- Marcin Rogawski implementation of Keccak round - based on C implementation
-- Marcin Rogawski implementation is smaller and faster on Altera (Stratix II-IV, Cyclone II-IV)
-- and Xilinx (Virtex 4-6, Spartan 3 and 6)	devices than Guido Bertoni implementation
entity keccak_fv8_round is
port (														
	clk		: in std_logic;
	
	selCI		: in std_logic;
	selTRP		: in std_logic;
	selLastByte : in std_logic;
	enSaveRotatedRho : in std_logic;
	byteAddr 	: in std_logic_vector(2 downto 0);
	
	rcCarry : in std_logic; -- the last bit of the 64-bit round constant in the next round 
    rin     : in std_logic_vector(199 downto 0); 	
    rc		: in std_logic_vector(7 downto 0);
    rout    : out std_logic_vector(199 downto 0));
end keccak_fv8_round;
			
architecture structure of keccak_fv8_round is
	
	signal aa : field_type;
	signal rhoIn : field_type;
	signal trp : field_type;	
	signal ciIn, ci : field_type;	
	signal ciMux, trpMux : field_type;	
	
	signal sum0, sum1, sum2, sum3, sum4: std_logic_vector(7 downto 0);
	signal fsum0, fsum1, fsum2, fsum3, fsum4: std_logic_vector(7 downto 0);
	
	signal preCalcSum : std_logic_vector(4 downto 0);
	
	signal sumCarry : std_logic_vector(4 downto 0);
	
	
	type sum_pre_calc_type is array ( 0 to 4 ) of std_logic_vector(4 downto 0);
	signal ss : sum_pre_calc_type;			   
	
	
	-- Note:
	-- _rot = temporary rotation register		 
	-- _hold = Rotation Hold (from first clock cycle of a round (byte 0) to usage (depending on the rotation constant, could be cycle 1 - 7 of the next round)	
	signal o01_rot, o01_hold : std_logic_vector(3 downto 0);    
	signal o02_rot	 		  : std_logic_vector(2 downto 0);    
	signal o03_rot, o03_hold : std_logic;            
	signal o04_rot, o04_hold : std_logic_vector(1 downto 0);    
	                                                  
	signal o10_rot	    	  : std_logic;                       
	signal o11_rot, o11_hold : std_logic_vector(3 downto 0);    
	signal o12_rot, o12_hold : std_logic_vector(1 downto 0);    
	signal o13_rot, o13_hold : std_logic_vector(4 downto 0);    
	signal o14_rot 		  : std_logic_vector(1 downto 0);    
	                                                  
	signal o20_rot, o20_hold : std_logic_vector(5 downto 0);    
	signal o21_rot           : std_logic_vector(5 downto 0);    
	signal o22_rot, o22_hold : std_logic_vector(2 downto 0);    
	signal o23_rot, o23_hold : std_logic_vector(6 downto 0);    
	signal o24_rot, o24_hold : std_logic_vector(4 downto 0);    
	                                                  
	signal o30_rot, o30_hold : std_logic_vector(3 downto 0);    
	signal o31_rot, o31_hold : std_logic_vector(6 downto 0);    
	signal o32_rot, o32_hold : std_logic;                       
	signal o33_rot, o33_hold : std_logic_vector(4 downto 0);    
	
	                                                  
	signal o40_rot, o40_hold : std_logic_vector(2 downto 0);    
	signal o41_rot, o41_hold : std_logic_vector(3 downto 0);    
	signal o42_rot, o42_hold : std_logic_vector(6 downto 0);    
									    
	signal o44_rot, o44_hold : std_logic_vector(5 downto 0);															
begin 														
	aa <= str2field(rin);		
																					 
	RotateRegister: process( clk )
	begin
		if rising_edge( clk ) then				  
			if selLastByte = '0' then
				sumCarry <= sum4(7) & sum3(7) & sum2(7) & sum1(7) & sum0(7);
			else
				sumCarry <= preCalcSum;
			end if;
		end if;
	end process;	
	
	-- Precalculation of carry sum for the first byte of the next block 5-bit total
	ss(0)(0) <= trp(0)(0)(7);
	ss(1)(2) <= trp(1)(2)(7); 
	ss(2)(0) <= trp(2)(0)(7);
	ss(2)(1) <= trp(2)(1)(7);
	ss(4)(4) <= trp(4)(4)(7); 
	
	RotateRegisterPrecalc: process( clk )
	begin
		if rising_edge( clk ) then
			--ss(0)(0) <= trp(0)(0)(7);  (Direct Pass Through)
			if byteAddr = 2 then	ss(0)(1) <= trp(0)(1)(7);	end if;
			if byteAddr = 2 then	ss(0)(2) <= trp(0)(2)(7);	end if;
			if byteAddr = 5 then	ss(0)(3) <= trp(0)(3)(7);	end if;
			if byteAddr = 6 then	ss(0)(4) <= trp(0)(4)(7);	end if;
			
			if byteAddr = 4 then	ss(1)(0) <= trp(1)(0)(7);	end if;
			if byteAddr = 5 then	ss(1)(1) <= trp(1)(1)(7);	end if;
			--ss(1)(2) <= trp(1)(2)(7);  (Direct Pass Through)
			if byteAddr = 2 then	ss(1)(3) <= trp(1)(3)(7);	end if;
			if byteAddr = 0 then	ss(1)(4) <= trp(1)(4)(7);	end if;
			
			--ss(2)(0) <= trp(2)(0)(7);  (Direct Pass Through)
			--ss(2)(1) <= trp(2)(1)(7);  (Direct Pass Through)
			if byteAddr = 4 then	ss(2)(2) <= trp(2)(2)(7);	end if;
			if byteAddr = 6 then	ss(2)(3) <= trp(2)(3)(7);	end if;
			if byteAddr = 5 then	ss(2)(4) <= trp(2)(4)(7);	end if;
			
			if byteAddr = 4 then	ss(3)(0) <= trp(3)(0)(7);	end if;
			if byteAddr = 3 then	ss(3)(1) <= trp(3)(1)(7);	end if;
			if byteAddr = 6 then	ss(3)(2) <= trp(3)(2)(7);	end if;
			if byteAddr = 6 then	ss(3)(3) <= trp(3)(3)(7);	end if;
			if byteAddr = 0 then	ss(3)(4) <= trp(3)(4)(7);	end if;
			
			if byteAddr = 0 then	ss(4)(0) <= trp(4)(0)(7);	end if;
			if byteAddr = 1 then	ss(4)(1) <= trp(4)(1)(7);	end if;
			if byteAddr = 3 then	ss(4)(2) <= trp(4)(2)(7);	end if;
			if byteAddr = 2 then	ss(4)(3) <= trp(4)(3)(7);	end if;
			--ss(4)(4) <= trp(4)(4)(7);  (Direct Pass Through)
		end if;
	end process;
	
	preCalcSum(0) <= 	(ss(0)(0) xor ((not ss(0)(1)) and ss(0)(2)) xor rcCarry) xor 	 -- (0)(0)						
						(ss(1)(0) xor ((not ss(1)(1)) and ss(1)(2))) xor 	 		 -- (1)(0)
						(ss(2)(0) xor ((not ss(2)(1)) and ss(2)(2))) xor 			 -- (2)(0)
						(ss(3)(0) xor ((not ss(3)(1)) and ss(3)(2))) xor 	 		 -- (3)(0)
						(ss(4)(0) xor ((not ss(4)(1)) and ss(4)(2)));	 			 -- (4)(0)
							  
	preCalcSum(1) <= 	(ss(0)(1) xor ((not ss(0)(2)) and ss(0)(3))) xor			 -- (0)(1)
						(ss(1)(1) xor ((not ss(1)(2)) and ss(1)(3))) xor 			 -- (1)(1)
						(ss(2)(1) xor ((not ss(2)(2)) and ss(2)(3))) xor 			 -- (2)(1)
						(ss(3)(1) xor ((not ss(3)(2)) and ss(3)(3))) xor 			 -- (3)(1)
						(ss(4)(1) xor ((not ss(4)(2)) and ss(4)(3)));	 			 -- (4)(1)
																											
	preCalcSum(2) <= 	(ss(0)(2) xor ((not ss(0)(3)) and ss(0)(4))) xor 			 -- (0)(2)
						(ss(1)(2) xor ((not ss(1)(3)) and ss(1)(4))) xor 			 -- (1)(2)						
						(ss(2)(2) xor ((not ss(2)(3)) and ss(2)(4))) xor 			 -- (2)(2)
						(ss(3)(2) xor ((not ss(3)(3)) and ss(3)(4))) xor 			 -- (3)(2)
						(ss(4)(2) xor ((not ss(4)(3)) and ss(4)(4)));	 			 -- (4)(2)
								  
	preCalcSum(3) <= 	(ss(0)(3) xor ((not ss(0)(4)) and ss(0)(0))) xor 	 		 -- (0)(3)
						(ss(1)(3) xor ((not ss(1)(4)) and ss(1)(0))) xor 			 -- (1)(3)
						(ss(2)(3) xor ((not ss(2)(4)) and ss(2)(0))) xor 			 -- (2)(3)						
						(ss(3)(3) xor ((not ss(3)(4)) and ss(3)(0))) xor 			 -- (3)(3)
						(ss(4)(3) xor ((not ss(4)(4)) and ss(4)(0)));	 			 -- (4)(3)
	
	preCalcSum(4) <= 	(ss(0)(4) xor ((not ss(0)(0)) and ss(0)(1))) xor			 -- (0)(4)
						(ss(1)(4) xor ((not ss(1)(0)) and ss(1)(1))) xor			 -- (1)(4)
						(ss(2)(4) xor ((not ss(2)(0)) and ss(2)(1))) xor			 -- (2)(4)
						(ss(3)(4) xor ((not ss(3)(0)) and ss(3)(1))) xor			 -- (3)(4)
						(ss(4)(4) xor ((not ss(4)(0)) and ss(4)(1)));				 -- (4)(4)
																 
	-- Sum Calc
	sum0 <= ciMux(0)(0) xor ciMux(1)(0) xor ciMux(2)(0) xor ciMux(3)(0) xor ciMux(4)(0);
	sum1 <= ciMux(0)(1) xor ciMux(1)(1) xor ciMux(2)(1) xor ciMux(3)(1) xor ciMux(4)(1);
	sum2 <= ciMux(0)(2) xor ciMux(1)(2) xor ciMux(2)(2) xor ciMux(3)(2) xor ciMux(4)(2);
	sum3 <= ciMux(0)(3) xor ciMux(1)(3) xor ciMux(2)(3) xor ciMux(3)(3) xor ciMux(4)(3);
	sum4 <= ciMux(0)(4) xor ciMux(1)(4) xor ciMux(2)(4) xor ciMux(3)(4) xor ciMux(4)(4);

	fsum0 <= sum4 xor sum1(6 downto 0) & sumCarry(1);
	fsum1 <= sum0 xor sum2(6 downto 0) & sumCarry(2);
	fsum2 <= sum1 xor sum3(6 downto 0) & sumCarry(3);
	fsum3 <= sum2 xor sum4(6 downto 0) & sumCarry(4);
	fsum4 <= sum3 xor sum0(6 downto 0) & sumCarry(0);
	
		
	rhoIn(0)(0) <= ciMux(0)(0) xor fsum0;	
	rhoIn(0)(1) <= ciMux(1)(0) xor fsum0; 	
	rhoIn(0)(2) <= ciMux(2)(0) xor fsum0; 	
	rhoIn(0)(3) <= ciMux(3)(0) xor fsum0; 	
	rhoIn(0)(4) <= ciMux(4)(0) xor fsum0; 	
	
	rhoIn(1)(0) <= ciMux(0)(1) xor fsum1; 	
	rhoIn(1)(1) <= ciMux(1)(1) xor fsum1; 	
	rhoIn(1)(2) <= ciMux(2)(1) xor fsum1; 	
	rhoIn(1)(3) <= ciMux(3)(1) xor fsum1; 	
	rhoIn(1)(4) <= ciMux(4)(1) xor fsum1;	
	
	rhoIn(2)(0) <= ciMux(0)(2) xor fsum2; 	
	rhoIn(2)(1) <= ciMux(1)(2) xor fsum2; 	
	rhoIn(2)(2) <= ciMux(2)(2) xor fsum2; 	
	rhoIn(2)(3) <= ciMux(3)(2) xor fsum2; 	
	rhoIn(2)(4) <= ciMux(4)(2) xor fsum2; 	
	
	rhoIn(3)(0) <= ciMux(0)(3) xor fsum3; 	
	rhoIn(3)(1) <= ciMux(1)(3) xor fsum3; 	
	rhoIn(3)(2) <= ciMux(2)(3) xor fsum3; 	
	rhoIn(3)(3) <= ciMux(3)(3) xor fsum3; 	
	rhoIn(3)(4) <= ciMux(4)(3) xor fsum3;	
	
	rhoIn(4)(0) <= ciMux(0)(4) xor fsum4; 	
	rhoIn(4)(1) <= ciMux(1)(4) xor fsum4; 	
	rhoIn(4)(2) <= ciMux(2)(4) xor fsum4; 	
	rhoIn(4)(3) <= ciMux(3)(4) xor fsum4; 	
	rhoIn(4)(4) <= ciMux(4)(4) xor fsum4; 	
 
	InterReg2: process( clk )
    begin
        if(rising_edge(clk)) then					 
			-- o00_rot <= rhoIn(0)(0);		-- unused
			o01_rot <= rhoIn(0)(1)(7 downto 4);
			o02_rot <= rhoIn(0)(2)(7 downto 5);
			o03_rot <= rhoIn(0)(3)(7);
			o04_rot <= rhoIn(0)(4)(7 downto 6);

            o10_rot <= rhoIn(1)(0)(7);		
			o11_rot <= rhoIn(1)(1)(7 downto 4);
			o12_rot <= rhoIn(1)(2)(7 downto 6);
			o13_rot <= rhoIn(1)(3)(7 downto 3);
			o14_rot <= rhoIn(1)(4)(7 downto 6);
			
            o20_rot <= rhoIn(2)(0)(7 downto 2);
			o21_rot <= rhoIn(2)(1)(7 downto 2);
			o22_rot <= rhoIn(2)(2)(7 downto 5);
			o23_rot <= rhoIn(2)(3)(7 downto 1);
			o24_rot <= rhoIn(2)(4)(7 downto 3);

			o30_rot <= rhoIn(3)(0)(7 downto 4);
			o31_rot <= rhoIn(3)(1)(7 downto 1);
			o32_rot <= rhoIn(3)(2)(7);
			o33_rot <= rhoIn(3)(3)(7 downto 3);
            -- o34_rot <= rhoIn(3)(4);		-- unused
			
            o40_rot <= rhoIn(4)(0)(7 downto 5);
			o41_rot <= rhoIn(4)(1)(7 downto 4);
			o42_rot <= rhoIn(4)(2)(7 downto 1);
            -- o43_rot <= rhoIn(4)(3);		-- unused
			o44_rot <= rhoIn(4)(4)(7 downto 2);
				
			if enSaveRotatedRho = '1' then
	           
				o01_hold <= o01_rot;
				
				o03_hold <= o03_rot;
				o04_hold <= o04_rot;
							       

				o11_hold <= o11_rot;
				o12_hold <= o12_rot;   
				o13_hold <= o13_rot;

							       
	            o20_hold <= o20_rot; 

				o22_hold <= o22_rot;
				o23_hold <= o23_rot;
				o24_hold <= o24_rot;
							       
				o30_hold <= o30_rot;
				o31_hold <= o31_rot;
				o32_hold <= o32_rot;
				o33_hold <= o33_rot;

							       
	            o40_hold <= o40_rot;
				o41_hold <= o41_rot;
				o42_hold <= o42_rot;

				o44_hold <= o44_rot;
			end if;
		end if;	
    end process;
	
	-- Perform 64-bit permutation here, byte rotation is done via selecting write address
	-- Note: Debugging TRP, looking at the state register should be easier to see as "write" location is involved.				   
   	trp(0)(0) <= rhoIn(0)(0);
	trp(3)(1) <= rhoIn(0)(1)(3 downto 0) & o01_rot;	
	trp(1)(2) <= rhoIn(0)(2)(4 downto 0) & o02_rot;	
	trp(4)(3) <= rhoIn(0)(3)(6 downto 0) & o03_rot;	
	trp(2)(4) <= rhoIn(0)(4)(5 downto 0) & o04_rot;	
				
	trp(2)(0) <= rhoIn(1)(0)(6 downto 0) & o10_rot;	     
	trp(0)(1) <= rhoIn(1)(1)(3 downto 0) & o11_rot;
	trp(3)(2) <= rhoIn(1)(2)(5 downto 0) & o12_rot;
	trp(1)(3) <= rhoIn(1)(3)(2 downto 0) & o13_rot;
	trp(4)(4) <= rhoIn(1)(4)(5 downto 0) & o14_rot;
				
	trp(4)(0) <= rhoIn(2)(0)(1 downto 0) & o20_rot;	     
	trp(2)(1) <= rhoIn(2)(1)(1 downto 0) & o21_rot;
	trp(0)(2) <= rhoIn(2)(2)(4 downto 0) & o22_rot;
	trp(3)(3) <= rhoIn(2)(3)(0) 		 & o23_rot;
	trp(1)(4) <= rhoIn(2)(4)(2 downto 0) & o24_rot;
				
	trp(1)(0) <= rhoIn(3)(0)(3 downto 0) & o30_rot;	     
	trp(4)(1) <= rhoIn(3)(1)(0) 		 & o31_rot;
	trp(2)(2) <= rhoIn(3)(2)(6 downto 0) & o32_rot;
	trp(0)(3) <= rhoIn(3)(3)(2 downto 0) & o33_rot;
	trp(3)(4) <= rhoIn(3)(4);					
				
	trp(3)(0) <= rhoIn(4)(0)(4 downto 0) & o40_rot;	     
	trp(1)(1) <= rhoIn(4)(1)(3 downto 0) & o41_rot;
	trp(4)(2) <= rhoIn(4)(2)(0) 		 & o42_rot;
	trp(2)(3) <= rhoIn(4)(3);
	trp(0)(4) <= rhoIn(4)(4)(1 downto 0) & o44_rot;	   
	
	
	-- CI in 
	-- Rotations are forwarded depending on the round number.
	ciIn(0)(0) <= aa(0)(0);
	ciIn(3)(1) <= aa(3)(1)(7 downto 4) & o01_hold when ( byteAddr = 4 ) else aa(3)(1);
	ciIn(1)(2) <= aa(1)(2)(7 downto 3) & o02_rot  when ( byteAddr = 0 ) else aa(1)(2);
	ciIn(4)(3) <= aa(4)(3)(7 downto 1) & o03_hold when ( byteAddr = 5 ) else aa(4)(3);
	ciIn(2)(4) <= aa(2)(4)(7 downto 2) & o04_hold when ( byteAddr = 2 ) else aa(2)(4);
				   		                                         		           
	ciIn(2)(0) <= aa(2)(0)(7 downto 1) & o10_rot  when ( byteAddr = 0 ) else aa(2)(0);
	ciIn(0)(1) <= aa(0)(1)(7 downto 4) & o11_hold when ( byteAddr = 5 ) else aa(0)(1);
	ciIn(3)(2) <= aa(3)(2)(7 downto 2) & o12_hold when ( byteAddr = 1 ) else aa(3)(2);
	ciIn(1)(3) <= aa(1)(3)(7 downto 5) & o13_hold when ( byteAddr = 5 ) else aa(1)(3);
	ciIn(4)(4) <= aa(4)(4)(7 downto 2) & o14_rot  when ( byteAddr = 0 ) else aa(4)(4);
				   		                                         		           
	ciIn(4)(0) <= aa(4)(0)(7 downto 6) & o20_hold when ( byteAddr = 7 ) else aa(4)(0);
	ciIn(2)(1) <= aa(2)(1)(7 downto 6) & o21_rot  when ( byteAddr = 0 ) else aa(2)(1);
	ciIn(0)(2) <= aa(0)(2)(7 downto 3) & o22_hold when ( byteAddr = 5 ) else aa(0)(2);
	ciIn(3)(3) <= aa(3)(3)(7)          & o23_hold when ( byteAddr = 1 ) else aa(3)(3);
	ciIn(1)(4) <= aa(1)(4)(7 downto 5) & o24_hold when ( byteAddr = 7 ) else aa(1)(4);
				   		                   		           
	ciIn(1)(0) <= aa(1)(0)(7 downto 4) & o30_hold when ( byteAddr = 3 ) else aa(1)(0);
	ciIn(4)(1) <= aa(4)(1)(7)		     & o31_hold when ( byteAddr = 6 ) else aa(4)(1);
	ciIn(2)(2) <= aa(2)(2)(7 downto 1) & o32_hold when ( byteAddr = 3 ) else aa(2)(2);
	ciIn(0)(3) <= aa(0)(3)(7 downto 5) & o33_hold when ( byteAddr = 2 ) else aa(0)(3);
	ciIn(3)(4) <= aa(3)(4);
				                                                		           
	ciIn(3)(0) <= aa(3)(0)(7 downto 3) & o40_hold when ( byteAddr = 3 ) else aa(3)(0);
	ciIn(1)(1) <= aa(1)(1)(7 downto 4) & o41_hold when ( byteAddr = 2 ) else aa(1)(1);
	ciIn(4)(2) <= aa(4)(2)(7) 		     & o42_hold when ( byteAddr = 4 ) else aa(4)(2);
	ciIn(2)(3) <= aa(2)(3);
	ciIn(0)(4) <= aa(0)(4)(7 downto 6) & o44_hold when ( byteAddr = 1 ) else aa(0)(4);


	-- CI
	ci(0)(0) <= (ciIn(0)(0) xor ((not ciIn(0)(1)) and ciIn(0)(2))) xor rc; 	
	ci(0)(1) <= ciIn(0)(1) xor ((not ciIn(0)(2)) and ciIn(0)(3)); 	
	ci(0)(2) <= ciIn(0)(2) xor ((not ciIn(0)(3)) and ciIn(0)(4));
	ci(0)(3) <= ciIn(0)(3) xor ((not ciIn(0)(4)) and ciIn(0)(0)); 	
	ci(0)(4) <= ciIn(0)(4) xor ((not ciIn(0)(0)) and ciIn(0)(1));
	
	ci(1)(0) <= ciIn(1)(0) xor ((not ciIn(1)(1)) and ciIn(1)(2)); 		
	ci(1)(1) <= ciIn(1)(1) xor ((not ciIn(1)(2)) and ciIn(1)(3));	
	ci(1)(2) <= ciIn(1)(2) xor ((not ciIn(1)(3)) and ciIn(1)(4));
	ci(1)(3) <= ciIn(1)(3) xor ((not ciIn(1)(4)) and ciIn(1)(0));
	ci(1)(4) <= ciIn(1)(4) xor ((not ciIn(1)(0)) and ciIn(1)(1));
	
	ci(2)(0) <= ciIn(2)(0) xor ((not ciIn(2)(1)) and ciIn(2)(2));
	ci(2)(1) <= ciIn(2)(1) xor ((not ciIn(2)(2)) and ciIn(2)(3));
	ci(2)(2) <= ciIn(2)(2) xor ((not ciIn(2)(3)) and ciIn(2)(4));
	ci(2)(3) <= ciIn(2)(3) xor ((not ciIn(2)(4)) and ciIn(2)(0));
	ci(2)(4) <= ciIn(2)(4) xor ((not ciIn(2)(0)) and ciIn(2)(1)); 	
	
	ci(3)(0) <= ciIn(3)(0) xor ((not ciIn(3)(1)) and ciIn(3)(2)); 	
	ci(3)(1) <= ciIn(3)(1) xor ((not ciIn(3)(2)) and ciIn(3)(3));
	ci(3)(2) <= ciIn(3)(2) xor ((not ciIn(3)(3)) and ciIn(3)(4));
	ci(3)(3) <= ciIn(3)(3) xor ((not ciIn(3)(4)) and ciIn(3)(0));
	ci(3)(4) <= ciIn(3)(4) xor ((not ciIn(3)(0)) and ciIn(3)(1));
	
	ci(4)(0) <= ciIn(4)(0) xor ((not ciIn(4)(1)) and ciIn(4)(2));
	ci(4)(1) <= ciIn(4)(1) xor ((not ciIn(4)(2)) and ciIn(4)(3));
	ci(4)(2) <= ciIn(4)(2) xor ((not ciIn(4)(3)) and ciIn(4)(4));
	ci(4)(3) <= ciIn(4)(3) xor ((not ciIn(4)(4)) and ciIn(4)(0));
	ci(4)(4) <= ciIn(4)(4) xor ((not ciIn(4)(0)) and ciIn(4)(1));
	
	ciMux <= ci when selCI = '1' else aa;
	--=====================================================
	trpMux <= trp when selTRP = '1' else ci;
	rout <= field2str(trpMux);
end;	


